-- This module is used to read data from LRB and forward them to the STC card
-- After reset, the base address of the LRB must be loaded before it can read data from the LRB. 
-- When the interface finds data available in LRB, it requests bus. After reset, the first word
-- from the LRB is marked as a header. When a disconnect with data is detected, that word is marked
-- as a trailer. Next data word will be marked as header again. Valid data sent to STC card is flaged
-- by stc_write. stc_data and flags are synchronized by stc_clk, which rising edge is norminally in
-- the middle of data and flag transitions.
-- LRB is not allowed to disconnect without data.
-- No error checking is implemented at this moment.
-- 1/4/01 Shouxiang Wu
-- 7/9/03 FRC readout disable added. no_trailer removed with latency timer disabled
--
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;

entity pci1_if is
    port (
-- pci clock and reset signals
        clk: in STD_LOGIC;
        rstn: in STD_LOGIC;
-- pci function local side signals
        stat_reg: in STD_LOGIC_VECTOR (5 downto 0);
        l_adro: in STD_LOGIC_VECTOR (31 downto 0);
        l_dato: in STD_LOGIC_VECTOR (31 downto 0);
        l_cmdo: in STD_LOGIC_VECTOR (3 downto 0);
        l_cbeni: out STD_LOGIC_VECTOR (3 downto 0);
        l_adi: out STD_LOGIC_VECTOR (31 downto 0);
-- pci function master local side signal
        lm_tsr: in STD_LOGIC_VECTOR (9 downto 0);
        lm_dxfrn: in STD_LOGIC;
        lm_adr_ackn: in STD_LOGIC;
        lm_ackn: in STD_LOGIC;
        lm_lastn: out STD_LOGIC;
        lm_rdyn: out STD_LOGIC;
        lm_req32n: out STD_LOGIC;
-- pci function target local side signal
        lt_framen: in STD_LOGIC;
        lt_dxfrn: in STD_LOGIC;
        lt_ackn: in STD_LOGIC;
        lt_tsr: in STD_LOGIC_VECTOR (11 downto 0);
        lt_discn: out STD_LOGIC;
        lt_rdyn: out STD_LOGIC;
-- data from/to stc
        stc_clk: in STD_LOGIC;
        stc_ctl_o: in STD_LOGIC_VECTOR (3 downto 0);
        stc_ctl_i: out STD_LOGIC_VECTOR (3 downto 0);
        stc_data: out STD_LOGIC_VECTOR (31 downto 0);
-- link receiver data available
        mb_a: in STD_LOGIC_VECTOR(4 downto 1);
        mb_b: in STD_LOGIC_VECTOR(4 downto 1);
        mb_c: in STD_LOGIC_VECTOR(4 downto 1)
    );
end pci1_if;

architecture pci1_arch of pci1_if is
COMPONENT lpm_ram_dp
   GENERIC ( LPM_WIDTH: POSITIVE :=32;
             LPM_WIDTHAD: POSITIVE := 8;
             LPM_NUMWORDS: NATURAL := 256;
             LPM_TYPE: STRING := "LPM_RAM_DP";
             LPM_INDATA: STRING := "REGISTERED";
             LPM_OUTDATA: STRING := "UNREGISTERED";
             LPM_RDADDRESS_CONTROL: STRING := "REGISTERED";
             LPM_WRADDRESS_CONTROL: STRING := "REGISTERED";
             LPM_FILE: STRING := "UNUSED";
             LPM_HINT: STRING := "UNUSED"

           );
   PORT    ( rdaddress, wraddress: IN STD_LOGIC_VECTOR(LPM_WIDTHAD-1 DOWNTO 0);
             rdclock, wrclock: IN STD_LOGIC := '0';
             rden, rdclken, wrclken: IN STD_LOGIC := '1';
             wren: IN STD_LOGIC; 
             data: IN STD_LOGIC_VECTOR(LPM_WIDTH-1 DOWNTO 0);
             q: OUT STD_LOGIC_VECTOR(LPM_WIDTH-1 DOWNTO 0));
END COMPONENT;
COMPONENT lpm_counter
   GENERIC (LPM_WIDTH: POSITIVE := 5;
      LPM_MODULUS: NATURAL := 0;
      LPM_DIRECTION: STRING := "UNUSED";
      LPM_AVALUE: STRING := "UNUSED";
      LPM_SVALUE: STRING := "UNUSED";
      LPM_PVALUE: STRING := "UNUSED";
      LPM_TYPE: STRING := "LPM_COUNTER";
      LPM_HINT: STRING := "UNUSED");
   PORT (data: IN STD_LOGIC_VECTOR(LPM_WIDTH-1 DOWNTO 0) := (OTHERS => '0');
      clock: IN STD_LOGIC;
      cin: IN STD_LOGIC := '0';
      clk_en: IN STD_LOGIC := '1';
      cnt_en: IN STD_LOGIC := '1';
      updown: IN STD_LOGIC := '1';
      sload: IN STD_LOGIC := '0';
      sset: IN STD_LOGIC := '0';
      sclr: IN STD_LOGIC := '0';
      aload: IN STD_LOGIC := '0';
      aset: IN STD_LOGIC := '0';
      aclr: IN STD_LOGIC := '0';
      cout: OUT STD_LOGIC := '0';
      q: OUT STD_LOGIC_VECTOR(LPM_WIDTH-1 DOWNTO 0));
END COMPONENT;
-- lrb base address, two LSBs are not implemented. Always read as zero.
signal lrb_addr: std_logic_vector(31 downto 0);
signal offset: std_logic_vector(14 downto 0);
signal tmp_data: std_logic_vector(31 downto 0);
signal fifo_din: std_logic_vector(32 downto 0);
signal fifo_dout: std_logic_vector(32 downto 0);
signal rts_cnt: std_logic_vector(1 downto 0);
signal RTS, lt_write, read_data, en_wr_fifo : std_logic;
signal fifo_ra,fifo_wa,event_no: std_logic_vector(7 downto 0);
signal fifo_wa_sync0,fifo_wa_sync1: std_logic_vector(2 downto 0);
signal msb: std_logic_vector(3 downto 0);
signal read_wait: unsigned(2 downto 0);
signal test_cntr: unsigned(7 downto 0);
signal test_header,fifo_re,en_master,inc_en: std_logic;
signal q0,q1,idle,header,trailer,xbit,stc_data_valid,fifo_data_valid,link_word,link_words,was_last_word: std_logic;
signal fifo_we,ec_fifo_wa,cin_wa,ec_fifo_ra,cin_ra,ld_fifo_ad,set_ad_h,ovfl: std_logic;
signal clr_ra,sel_tsr0,sel_fifo,test_start,test_start_r,test,test_cycle,reset: std_logic;
signal fifo_out_rdy: std_logic;
-- this avoids reading lrb forever
signal cnt_lrb_wc : std_logic;
signal rst_lrb_wc : std_logic;
signal lrb_wc_ovfl : std_logic;
-- lrb monitor data signals
signal rd_monitor,mon_data_out,mon_latch,mon_active,mfifo_we,mfifo_re: std_logic;
signal mon_data_ready,rst_active,start_bit,start_r,stop_shift,mfifo_dav: std_logic;
signal shift_data,start,mon_rqst,lm_last: std_logic;
signal lm_last_dl: std_logic_vector(1 downto 0);
signal mfifo_wa,mfifo_ra: std_logic_vector(3 downto 0);
signal mfifo_dout,mon_data: std_logic_vector(31 downto 0);
signal mon_active_sync: std_logic_vector(1 downto 0);
signal mon_data_ready_sync: std_logic_vector(1 downto 0);
signal shift_cntr: unsigned(4 downto 0);
begin
lm_lastn <= not lm_last;
process(stc_clk)
begin
  if(stc_clk'event and stc_clk = '1')then
    stc_ctl_i(0) <= stc_data_valid;
    stc_ctl_i(1) <= trailer or (not test and fifo_dout(32));
    stc_ctl_i(2) <= mon_data_out;
    stc_ctl_i(3) <= '0';
    stc_data(31 downto 8) <= fifo_dout(31 downto 8);
    if(test_header = '1' and mon_data_out = '0')then
      stc_data(7 downto 0) <= event_no;
    else
      stc_data(7 downto 0) <= fifo_dout(7 downto 0);
    end if;
    reset <= stc_ctl_o(0);
    test_start <= stc_ctl_o(1) and stc_ctl_o(2);
    test <= stc_ctl_o(2);
    mon_latch <= stc_ctl_o(3);
  end if;
end process;
-- 256 words of 32 bit fifo. top 8 words are used for pci clock to stc clock synchronization, bottom 248 words
-- are used as FRC test data memory
-- data format of test data is the same as FRC data, except a link word is added before the header word which contains
-- the word count of the FRC data in bit 7 thru 0. After the last block of FRC data, bit 31 of this word should be set
-- to '1' to force the address jump back to 0.
i_event_no: lpm_counter
   generic map(LPM_WIDTH => 8)
   port map(
           clock => stc_clk,
           cnt_en => inc_en,
           aclr => reset,
           q => event_no);
inc_en <= stc_data_valid and trailer and not mon_data_out;
fifo: lpm_ram_dp
   generic map(LPM_WIDTH => 33)
   port map(
           rdaddress => fifo_ra,
           wraddress => fifo_wa,
           wrclock => clk,
           rdclock => stc_clk,
           rden => fifo_re,
           wren => fifo_we,
           data => fifo_din,
           q => fifo_dout);
fifo_din <= not lm_tsr(3) & tmp_data;
fifo_re <= ec_fifo_ra or lt_tsr(1);
fifo_we <= (en_wr_fifo and (not lm_dxfrn or (read_data and not lm_tsr(3)))) or lt_write;
process(clk)
begin
  if(clk'event and clk = '1')then
    if(lm_dxfrn = '0' or lt_dxfrn = '0')then
      tmp_data <= l_dato;
    end if;
    en_wr_fifo <= lm_adr_ackn and (en_wr_fifo or (not lm_dxfrn and not rd_monitor));
    lt_write <= lt_tsr(1) and l_cmdo(0) and not lt_dxfrn;
  end if;
end process;
-- five MSBs of fifo write address
fifo_wa_h: lpm_counter
   port map(
           data => l_adro(9 downto 5),
           clock => clk,
           clk_en => ec_fifo_wa,
           cnt_en => cin_wa,
           aload => ld_fifo_ad,
           aset => set_ad_h,
           cout => ovfl,
           q => fifo_wa(7 downto 3));
cin_wa <= '1' when fifo_wa(2 downto 0) = "100" else '0';
-- three LSBs of fifo write address
process(clk,ld_fifo_ad)
begin
  if(ld_fifo_ad = '1')then
    fifo_wa(0) <= l_adro(3) xor l_adro(2);
    fifo_wa(1) <= l_adro(4) xor l_adro(3);
    fifo_wa(2) <= l_adro(4);
  elsif(clk'event and clk = '1')then
    if(ec_fifo_wa = '1')then
      fifo_wa(2) <= (fifo_wa(0) and fifo_wa(2)) or (not fifo_wa(0) and fifo_wa(1));
      fifo_wa(1) <= (fifo_wa(0) and not fifo_wa(2)) or (not fifo_wa(0) and fifo_wa(1));
      fifo_wa(0) <= not fifo_wa(2) xor fifo_wa(1);
    end if; 
  end if;
end process;
-- five MSBs of fifo read address
fifo_ra_h: lpm_counter
   port map(
           data => fifo_wa(7 downto 3),
           clock => stc_clk,
           cnt_en => cin_ra,
           aload => lt_tsr(1),
           aset => set_ad_h,
           sclr => clr_ra,
           q => fifo_ra(7 downto 3));
cin_ra <= '1' when fifo_ra(2 downto 0) = "100" and test_cycle = '1' else '0';
clr_ra <= (link_word and fifo_dout(31)) or (reset and test);
-- three LSBs of fifo read address
process(stc_clk,lt_tsr)
begin
  if(lt_tsr(1) = '1')then
    fifo_ra(2 downto 0) <= fifo_wa(2 downto 0);
    fifo_wa_sync0 <= "000";
    fifo_wa_sync1 <= "000";
  elsif(stc_clk'event and stc_clk = '1')then
    if(clr_ra = '1')then
      fifo_ra(2 downto 0) <= "000";
    elsif(ec_fifo_ra = '1')then
      fifo_ra(2) <= (fifo_ra(0) and fifo_ra(2)) or (not fifo_ra(0) and fifo_ra(1));
      fifo_ra(1) <= (fifo_ra(0) and not fifo_ra(2)) or (not fifo_ra(0) and fifo_ra(1));
      fifo_ra(0) <= not fifo_ra(2) xor fifo_ra(1);
    end if;
    fifo_wa_sync0 <= fifo_wa(2 downto 0);
    fifo_wa_sync1 <= fifo_wa_sync0;
  end if;
end process;
set_ad_h <= not (lt_tsr(1) or sel_fifo or test);
ld_fifo_ad <= lt_tsr(1) and not sel_fifo;
ec_fifo_wa <= (lt_tsr(1) and not l_cmdo(0) and not lt_dxfrn) or fifo_we;
ec_fifo_ra <= RTS when test_cycle = '1' or ( test = '0' and fifo_wa_sync1 /= fifo_ra(2 downto 0)) else '0';
stc_data_valid <= fifo_data_valid and not link_words;
process(clk,rstn)
begin
  if(rstn = '0')then
    sel_fifo <= '0';
  elsif(clk'event and clk = '1')then
    sel_fifo <= lt_tsr(1);
  end if;
end process;
process(stc_clk,reset)
begin
  if(reset = '1')then
    RTS <= '1';
    rts_cnt <= "00";
    header <= '1';
    test_header <= test;
    trailer <= '0';
    fifo_data_valid <= '0';
  elsif(stc_clk'event and stc_clk = '1')then
-- this inserts 4 idle cycles after the trailer to comply with STC pci1 interface requirement
    RTS <= (RTS and not (trailer or (not test and fifo_dout(32) and fifo_data_valid))) or (rts_cnt(1) and rts_cnt(0));
    rts_cnt(1) <= (rts_cnt(1) xor rts_cnt(0)) and not RTS;
    rts_cnt(0) <= not rts_cnt(0) and not RTS;
    fifo_data_valid <= ec_fifo_ra;
    if(test_cycle = '1' and test_cntr = 3)then
      trailer <= '1';
    else
      trailer <= '0';
    end if;
    if(trailer = '1' or (link_word = '1' and fifo_dout(31) = '0'))then
      test_header <= test and lrb_addr(2);
      header <= '1';
    elsif(stc_data_valid = '1')then
      test_header <= '0';
      header <= '0';
    end if;
  end if;
end process;
process(stc_clk,reset)
begin
  if(reset = '1')then
    test_cycle <= '0';
    test_start_r <= '0';
    test_cntr <= (others => '0');
    link_word <= '0';
    link_words <= '0';
    was_last_word <= '0';
  elsif(stc_clk'event and stc_clk = '1')then
    if(test_start = '1')then
      test_cycle <= '1';
    elsif(test_cntr = 3)then
      test_cycle <= link_words;
    end if;
    test_start_r <= test_start;
    if(link_words = '1')then
      for i in test_cntr'range loop
        test_cntr(i) <= fifo_dout(i);
      end loop;
    elsif(test_cycle = '1')then
      test_cntr <= test_cntr - 1;
    end if;
    was_last_word <= link_word and fifo_dout(31);
    link_words <= test_start_r or (link_words and not(link_word and not fifo_dout(31))); 
    link_word <= test_start_r or was_last_word;
  end if;
end process;
lt_rdyn <= not sel_tsr0 and not (sel_fifo and (l_cmdo(0) or fifo_out_rdy));
lt_discn <= not (sel_tsr0 and  not lt_dxfrn) and not(sel_fifo and ovfl and not lt_dxfrn);
process(clk,rstn,lrb_addr)
variable dav_sel: std_logic_vector(3 downto 0);
begin
  dav_sel := lrb_addr(1 downto 0) & lrb_addr(13 downto 12);
  if (rstn = '0')then
    sel_tsr0 <= '0';
    read_data <= '0'; 
    fifo_out_rdy <= '0';
    read_wait <= "000";
    q0 <= '0';
    q1 <= '0';
    en_master <= '0';
    lrb_addr <= (others => '0');
  elsif(clk'event and clk = '1')then
    sel_tsr0 <= lt_tsr(0);
    read_data <= lm_tsr(3);
    if(read_wait = 4)then
      fifo_out_rdy <= '1';
    elsif(lt_ackn = '0')then
      fifo_out_rdy <= '0';
    end if;
    if(sel_fifo = '0' or lt_dxfrn = '0')then
      read_wait <= "000";
    else
      read_wait <= read_wait + 1;
    end if;
-- target function has only one 32 bit register mapped to memory space, disconnect after single word access
-- It saves the LRB's base address
    if(lt_dxfrn = '0' and l_cmdo(0) = '1' and lt_tsr(0) = '1')then
      lrb_addr <= l_dato;
      en_master <= '1';
    end if;
-- master logic follows
-- lrb_dav is synchronized and used to generate bus request
    case dav_sel is
      when "0100" => q0 <= mb_a(1);
      when "0101" => q0 <= mb_a(2);
      when "0110" => q0 <= mb_a(3);
      when "1000" => q0 <= mb_b(1);
      when "1001" => q0 <= mb_b(2);
      when "1010" => q0 <= mb_b(3);
      when "1100" => q0 <= mb_c(1);
      when "1101" => q0 <= mb_c(2);
      when "1110" => q0 <= mb_c(3);
      when others => q0 <= '0';
    end case;
-- not in test mode, stc_ctl_o(1) is FRC readout disbale
    q1 <= q0 and not(stc_ctl_o(1) and not stc_ctl_o(2));
  end if;
end process;
i_lrb_wc: lpm_counter
   generic map(LPM_WIDTH => 6)
   port map(
           clock => clk,
           cnt_en => cnt_lrb_wc,
           aclr => rst_lrb_wc,
           cout => lrb_wc_ovfl);
rst_lrb_wc <= not lm_adr_ackn;
cnt_lrb_wc <= not lm_dxfrn;
-- master is always ready when in the data phase 
lm_rdyn <= not (lm_tsr(2) or lm_tsr(3));
lm_last <= not lm_dxfrn and ((rd_monitor and mfifo_wa(3) and mfifo_wa(0)) or lrb_wc_ovfl);
--lm_lastn <= not (lm_tsr(2) and rd_monitor);
-- if not in any of the active phases, master is in idle
idle <= not (lm_tsr(0) or lm_tsr(1) or lm_tsr(2) or lm_tsr(3)) and en_master;
-- request bus when idle and line receiver has data
lm_req32n <= not (((q1 and not test) or rd_monitor) and idle);
-- read multiple command and all bytes enabled
l_cbeni <= "0110" when (lm_adr_ackn = '0') else "1111";
-- 
offset(1 downto 0) <= lrb_addr(1 downto 0) when lt_tsr(0) = '1' else "00";
offset(14 downto 2) <= lrb_addr(14 downto 2) when rd_monitor = '0' else "0000000010010";
l_adi <= lrb_addr(31 downto 15) & offset when lt_tsr(1) = '0' else fifo_dout(31 downto 0);
--l_adi <= lrb_addr(31 downto 15) & offset when lt_tsr(1) = '0' else "00000000000000000000" & mb_a & mb_b & mb_c;
-- LRB monitor data logic follows 
process(clk,rstn)
begin
  if (rstn = '0')then
    mon_active_sync <= "00";
    mon_rqst <= '0';
    rd_monitor <= '0';
    lm_last_dl <= "00";
  elsif(clk'event and clk = '1')then
    mon_active_sync(0) <= mon_active;
    mon_active_sync(1) <= mon_active_sync(0);
    mon_rqst <= (mon_rqst and not rd_monitor) or (not mon_active_sync(1) and mon_active_sync(0));
    rd_monitor <= (rd_monitor and not lm_last_dl(1) and not(stat_reg(2) or stat_reg(3))) or (mon_rqst and idle);
    lm_last_dl <= lm_last_dl(0) & lm_last;
  end if;
end process;
mfifo_we <= not lm_dxfrn and rd_monitor;
process(stc_clk,reset)
begin
  if(reset = '1')then
    mon_active <= '0';
  elsif(stc_clk'event and stc_clk = '1')then
    mon_active <= en_master and (mon_latch or (stc_data_valid and fifo_dout(23) and header) or (mon_active and not rst_active));
  end if;
end process;
mon_fifo: lpm_ram_dp
   generic map (LPM_WIDTHAD => 4, LPM_NUMWORDS => 16)
   port map(
           rdaddress => mfifo_ra,
           wraddress => mfifo_wa,
           wrclock => clk,
           rdclock => stc_clk,
           rden => mfifo_re,
           wren => mfifo_we,
           data => l_dato,
           q => mfifo_dout);
process(clk,mon_active)
begin
  if(mon_active = '0')then
    mfifo_wa <= "0000";
    mon_data_ready <= '0';
  elsif(clk'event and clk = '1')then
    if(mfifo_we = '1')then
      mfifo_wa(3) <= mfifo_wa(3) xor (mfifo_wa(2) and mfifo_wa(1) and mfifo_wa(0));
      mfifo_wa(2) <= mfifo_wa(2) xor (mfifo_wa(1) and mfifo_wa(0));
      mfifo_wa(1) <= mfifo_wa(1) xor mfifo_wa(0);
      mfifo_wa(0) <= not mfifo_wa(0);
    end if;
    mon_data_ready <= (mfifo_wa(3) and mfifo_wa(0) and not lm_dxfrn) or mon_data_ready;
  end if;
end process;
process(stc_clk,mon_active)
begin
  if(mon_active = '0')then
    mfifo_ra <= "0000";
    mon_data_ready_sync <= "00";
    start <= '0';
    start_bit <= '0';
    stop_shift <= '0';
    rst_active <= '0';
    mfifo_re <= '0';
    mfifo_dav <= '0';
    shift_data <= '0';
    mon_data_out <= '0';
    mon_data <= (others => '0');
	shift_cntr <= "00000";
  elsif(stc_clk'event and stc_clk = '1')then
    if(mfifo_re = '1')then
      mfifo_ra(3) <= mfifo_ra(3) xor (mfifo_ra(2) and mfifo_ra(1) and mfifo_ra(0));
      mfifo_ra(2) <= mfifo_ra(2) xor (mfifo_ra(1) and mfifo_ra(0));
      mfifo_ra(1) <= mfifo_ra(1) xor mfifo_ra(0);
      mfifo_ra(0) <= not mfifo_ra(0);
    end if;
    mon_data_ready_sync(0) <= mon_data_ready;
    mon_data_ready_sync(1) <= mon_data_ready_sync(0);
    start <= not mon_data_ready_sync(1) and mon_data_ready_sync(0);
    if(shift_data = '1')then
      shift_cntr <= shift_cntr + 1;
    end if;
    if(shift_cntr = 30 or start = '1')then
      stop_shift <= mfifo_ra(3) and mfifo_ra(1);
      mfifo_re <= not(mfifo_ra(3) and mfifo_ra(1));
    else
      stop_shift <= '0';
      mfifo_re <= '0';
    end if;
    rst_active <= stop_shift; 
    mfifo_dav <= mfifo_re;
    shift_data <= mfifo_re or (shift_data and not stop_shift); 
    if(mfifo_dav = '1')then
      mon_data <= mfifo_dout;
    elsif(shift_data = '1')then
      mon_data <= '0' & mon_data(31 downto 1);
    end if;
    start_r <= start;
    start_bit <= start_r;
    mon_data_out <= start_bit or mon_data(0);
  end if;
end process;
end pci1_arch;
